package org.postgresql;


/* 
 * Author: Ned Wolpert <wolpert@yahoo.com>  
 * For Use By: Anyone under the postgresql license
 */

import java.io.PrintWriter;
import java.rmi.Remote;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Properties;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import javax.naming.Referenceable;
import javax.naming.Reference;
import javax.naming.StringRefAddr;
import javax.naming.RefAddr;
import javax.naming.Context;
import javax.naming.Name;
import javax.naming.NamingException;
import javax.naming.spi.ObjectFactory;
import javax.sql.ConnectionPoolDataSource;
import javax.sql.ConnectionEventListener;
import javax.sql.ConnectionEvent;
import javax.sql.PooledConnection;
import javax.sql.DataSource;
import org.postgresql.Field;
import org.postgresql.jdbc2.PostgresqlPooledConnection;
import org.postgresql.util.DefaultProperties;
import org.postgresql.util.PSQLException;


/**
 * Note, this was modified from the PostgresqlDataSource.  It provides
 * pooling for a limited number of connections, according to the JDBC2
 * optional specification.
 *
 * TODO: Fix Naming stuff. Cleaning thread to autoClose connections left open. Re-add-in ability to
 * keep a minimum number of connections 'free'.
 */

public class PostgresqlConnectionPoolDataSource implements ConnectionPoolDataSource, ConnectionEventListener,DataSource {
    
    // pooled behavior

    /**
     * Minimum number of connections to keep alives. (1 tends to be fine)
     */
    private int minNumberOfConnections=1; // default

    /**
     * Minimum number of connections to keep unused. (1 is usaully good enough)
     * FIXME: This is not currently used.  When the cleaner thread is implemented,
     * then this will be used.
     */
    private int minNumberOfConnectionsUnused=1;

    /**
     * Max number of connections to allow open.  (Should be bigger than min,
     * of course.)
     */
    private int maxNumberOfConnections=10;

    /**
     * Number of seconds before unused connections are closed
     * FIXME: This is not currently used.  When the cleaner thread is implemented,
     * then this will be used.
     */
    private int numberOfSecondsBeforeAutoClosed=300; // five min, eh?

    /**
     * Discard connections that had errors
     */
    private boolean discardConnectionsWithExceptions=true;

    // connection stuff below
    

	/**
	 * Holds the timeout for opening a new connection, specified
	 * in seconds. The default is obtained from the JDBC driver.
	 */
	private int _loginTimeout;


	/**
	 * Holds the user's account name.
	 */
	private String _user;


	/**
	 * Holds the database password.
	 */
	private String _password;


	/**
	 * Holds the name of the particular database on the server.
	 */
	private String _databaseName;


	/**
	 * Description of this datasource.
	 */
	private String _description = "PostgreSQL Connection Pooled DataSource";


	/**
	 * Holds the database server name. If null, this is
	 * assumed to be the localhost.
	 */
	private String _serverName;


	/**
	 * Holds the port number where a server is listening.
	 * The default value will open a connection with an
	 * unspecified port.
	 */
	private int _portNumber = DEFAULT_PORT;


	/**
	 * The default port number. Since we open the connection
	 * without specifying the port if it's the default one,
	 * this value can be meaningless.
	 */
	private static final int DEFAULT_PORT = 0;


	/**
	 * Holds the log writer to which all messages should be
	 * printed. The default writer is obtained from the driver
	 * manager, but it can be specified at the datasource level
	 * and will be passed to the driver. May be null.
	 */
	private transient PrintWriter _logWriter;

	/**
	 * Each datasource maintains it's own driver, in case of
	 * driver-specific setup (e.g. pools, log writer).
	 */
	private transient org.postgresql.Driver _driver;

    /**
     * Used for locking...
     */
    private transient Object LOCK = new Object();

    /**
     * Default constructor.  Note that the connections are not
     * initilized until the first request comes in.
     */
    public PostgresqlConnectionPoolDataSource(){
		_logWriter = DriverManager.getLogWriter();
		_loginTimeout = DriverManager.getLoginTimeout();
        readProperties();
	}


    /**
     * Returns a pooled connection based on the default username/password.
     * This is used to mimik a normal DataSource implementation.
     * <br>
     * @returns Connection
     * @throws SQLException should something fail.
     */
    public java.sql.Connection getConnection() throws SQLException{
        return getConnection(_user,_password);
    }


    /**
     * Returns a pooled connection based on the username/password given.
     * This is used to mimik a normal DataSource implementation.
     * <br>
     * @returns Connection
     * @throws SQLException should something fail.
     */
    public java.sql.Connection getConnection(String user, String password) throws SQLException{
        return (java.sql.Connection)getPooledConnection(user,password);
    }

    /**
     * Returns a pooled connection based on the default username/password.
     * <br>
     * @returns PooledConnection
     * @throws SQLException should something fail.
     */
    public PooledConnection getPooledConnection() throws SQLException{
        return getPooledConnection(_user,_password);
    }
    
    /**
     * Returns a pooled connection based on the username/password given.
     * <br>
     * Question: Currently, we just reset if the username/password
     * changes on us.  Maybe we need multiple pools we manage?
     *
     * @params String username to override the default
     * @params String password to override the default
     * @returns PooledConnection
     * @throws SQLException should something fail.
     */
    public PooledConnection getPooledConnection(String user, String password) throws SQLException {
        synchronized(LOCK){
            try {
                // if no connections, or username changed, we cause a reset
                if ( (connectionsUsed==null) ||
                     ( hasChanged(user,_user) || hasChanged(password,_password))){
                    // set the passwords just in case
                    setUser(user);
                    setPassword(password);
                    resetConnections();
                }
                // is a freeconnection available?
                return getUserConnection(); // Note, this method will block
                // if we are waiting for a connection
            }catch(SQLException e){
                print( "getPooledConnection() get Connection failed " + e );
                throw e;
            } finally {
                LOCK.notifyAll(); // we do this just to make sure other blockers are waken up
            }
        }
    }

    /**
     * Used to get a string representation of the queue status
     */
    private final String getQueueStatus(){
        StringBuffer sb = new StringBuffer("Free=");
        sb.append(connectionsUnused.size());
        sb.append(" Used="); 
        sb.append(connectionsUsed.size());
        return sb.toString();
    }

    /**
     * This is used to get a free pooled connection.  It will block
     * until the _loginTimeout period has expired.  This should be
     * called by a synchronized method to work properly.  <b>Note:</b>
     * This method is one of two that manipulate the list of used and
     * unused connections. (putUserConnection() is the other one)
     */
    protected PostgresqlPooledConnection getUserConnection() throws SQLException{
        // First, if no free connection is availalbe, and we have hit the max number
        // of connections, then we wait until the above is no longer true.  Our wait
        // interval is until we get notified, or half a second.  After a total of
        // of our _loginTimeout has passed, then we toss our SQLException
        long startTime = new java.util.Date().getTime();
        while ( (connectionsUnused.size()< 1) && 
                (connectionsUsed.size()  >= maxNumberOfConnections)) { 
            long now = new java.util.Date().getTime();
            int seconds = (int) ((now - startTime)/1000);
            if ( (_loginTimeout>0) && (seconds > _loginTimeout)){
                String queueStatus = getQueueStatus();
                print("getUserConnection() Login timeout execeeded: " + _loginTimeout + 
                      " "+ queueStatus);
                throw new SQLException("Login timeout exceeded " + queueStatus);
            }
            try {
                print("getUserConnection() Waiting for free connection. ");
                LOCK.wait(500);
            } catch (Exception e){
                ; // we don't care
            }
        }
        PostgresqlPooledConnection ppc = null;
        if (connectionsUnused.size()>0){ // use the free one
            print("getUserConnection() Using free connection");
            ppc = (PostgresqlPooledConnection) connectionsUnused.remove(0);
        } else { // we can generate a new one.
            print("getUserConnection() Creating new connection");
            ppc=_getPooledConnection();
        }
        connectionsUsed.add(ppc);
        print("getUserConnection() Returning connection to user. ");
        return ppc;
    }

    /**
     * Provides the ability to put back a used connection.  Errored
     * connection are dropped if 'discardConnectionsWithExceptions' is
     * set to true.  It does call synchronize to block on the LOCK
     * object. <b>Note:</b> This method is one of two that manipulate
     * the list of used and unused connections. (getUserConnection()
     * is the other one)
     */
    protected void putUserConnection(PostgresqlPooledConnection ppc) throws SQLException{
        synchronized(LOCK){
            if (!connectionsUsed.remove(ppc)){
                String queueStatus = getQueueStatus();
                boolean inUnused = connectionsUnused.contains(ppc);
                if (inUnused){
                    // then this was likely called for the connection and the pooled
                    // connection. Shouldn't be a problem, but bad programming so
                    // we print the error message for tracking purposes.
                    print("putUserConnection() Connection was closed when it wasn't open. " +
                          ppc);
                    return;
                } else { // Connection is totally lost.  Whoops.
                    print("putUserConnection() Lost connection: " + ppc + 
                          " does Unused contaion this: " + inUnused +
                          " " + queueStatus);
                    throw new SQLException("Lost Connection: " + ppc);
                }
            }
            if (ppc.hasException() && discardConnectionsWithExceptions){
                print("putUserConnection() Connection had error. Removing it. " + ppc);
                removeConnection(ppc);
                ppc = null;
            }
            // make sure that we need to keep this connection.
            // if we need more unused connetions
            // or we are at our minimum
            // then keep this connection (Or remade it if we destroyed it
            if ( (connectionsUnused.size() < minNumberOfConnections) ){
                if (ppc==null){
                    ppc = _getPooledConnection();
                    print("putUserConnection()  Replacing error connection with new connection " + ppc);
                }
                // do we need to double check that we aren't at the max?
                // logically, I think its covered by default
            } else { // drop the connection, we don't need it
                if (ppc!=null){
                    print("putUserConnection()  Discarding extra connection: " + ppc + 
                          " " + getQueueStatus());
                    removeConnection(ppc);
                    ppc=null;
                }
            }
            if (ppc!=null){ // then we want to re-add connection to our list
                connectionsUnused.add(ppc);
            }
            LOCK.notifyAll();
        }
    }

    /**
     * Wrapping the print here to make it consistant.
     */
    private final void print(String str){
        if (_logWriter!=null){
            _logWriter.println("[ JDBC-Pooler: " + Thread.currentThread()+ "  ] " + str);
        }
    }

    public PrintWriter getLogWriter()throws SQLException{
        return _logWriter;
    }

    public void setLogWriter(PrintWriter out)throws SQLException{
        _logWriter = out;
    }

    public void setLoginTimeout(int seconds) throws SQLException{
        _loginTimeout = seconds;
    }

    public int getLoginTimeout() throws SQLException{
        return _loginTimeout;
    }

    
    /**
     * This is how we tell if a connection has clossed.  We get the event here.
     */
    public void connectionClosed(ConnectionEvent event){
        try {
            putUserConnection((PostgresqlPooledConnection)event.getSource());
        } catch (SQLException e){
            print( "connectionClosed() Connection Close failed " + e );
        }
    }

    /**
     * Log the exception if the logger is present.
     */
    public void connectionErrorOccurred(ConnectionEvent event){
        print( "connectionErrorOccurred() " + event.getSQLException() );
    }



	/**
	 * Sets the name of the particular database on the server.
	 * The standard name for this property is <tt>databaseName</tt>.
	 *
	 * @param databaseName The name of the particular database on the server
	 */
	public synchronized void setDatabaseName( String databaseName )
	{
		if ( databaseName == null )
			throw new NullPointerException( "DataSource: Argument 'databaseName' is null" );
		_databaseName = databaseName;
	}


	/**
	 * Returns the name of the particular database on the server.
	 * The standard name for this property is <tt>databaseName</tt>.
	 *
	 * @return The name of the particular database on the server
	 */
	public String getDatabaseName()
	{
		return _databaseName;
	}


	/**
	 * Sets the description of this datasource.
	 * The standard name for this property is <tt>description</tt>.
	 *
	 * @param description The description of this datasource
	 */
	public synchronized void setDescription( String description )
	{
		if ( description == null )
			throw new NullPointerException( "DataSource: Argument 'description' is null" );
		_description = description;
	}


	/**
	 * Returns the description of this datasource.
	 * The standard name for this property is <tt>description</tt>.
	 *
	 * @return The description of this datasource
	 */
	public String getDescription()
	{
		return _description;
	}


	/**
	 * Sets the database password.
	 * The standard name for this property is <tt>password</tt>.
	 *
	 * @param password The database password
	 */
	public synchronized void setPassword( String password )
	{
		_password = password;
	}


	/**
	 * Returns the database password.
	 * The standard name for this property is <tt>password</tt>.
	 *
	 * @return The database password
	 */
	public String getPassword()
	{
		return _password;
	}


	/**
	 * Sets the port number where a server is listening.
	 * The standard name for this property is <tt>portNumber</tt>.
	 *
	 * @param portNumber The port number where a server is listening
	 */
	public synchronized void setPortNumber( int portNumber )
	{
		_portNumber = portNumber;
	}


	/**
	 * Returns the port number where a server is listening.
	 * The standard name for this property is <tt>portNumber</tt>.
	 *
	 * @return The port number where a server is listening
	 */
	public int getPortNumber()
	{
		return _portNumber;
	}


	/**
	 * Sets the database server name.

	 * The standard name for this property is <tt>serverName</tt>.
	 *
	 * @param serverName The database server name
	 */
	public synchronized void setServerName( String serverName )
	{
		_serverName = serverName;
	}


	/**
	 * Returns the database server name.
	 * The standard name for this property is <tt>serverName</tt>.
	 *
	 * @return The database server name
	 */
	public String getServerName()
	{
		return _serverName;
	}


	/**
	 * Sets the user's account name.
	 * The standard name for this property is <tt>user</tt>.
	 *
	 * @param user The user's account name
	 */
	public synchronized void setUser( String user )
	{
		_user = user;
	}


	/**
	 * Returns the user's account name.
	 * The standard name for this property is <tt>user</tt>.
	 *
	 * @return The user's account name
	 */
	public String getUser()
	{
		return _user;
	}


	/**
	 * Returns true if this datasource and the other are equal.
	 * The two datasources are equal if and only if they will produce
	 * the exact same connections. Connection properties like database
	 * name, user name, etc are comapred. Setup properties like
	 * description, log writer, etc are not compared.
	 */
	public synchronized boolean equals( Object other )
	{
		if ( other == this )
			return true;
		if ( other == null || ! ( other instanceof PostgresqlConnectionPoolDataSource ) )
			return false;

		PostgresqlConnectionPoolDataSource with;

		with = (PostgresqlConnectionPoolDataSource) other;
		if ( _databaseName != null && _databaseName.equals( with._databaseName ) )
			if ( _portNumber == with._portNumber &&
					( ( _serverName == null && with._serverName == null ) ||
					  ( _serverName != null && _serverName.equals( with._serverName ) ) ) )
				if ( ( _user == null && with._user == null ) ||
						( _user != null && _password != null && _user.equals( with._user ) &&
						  _password.equals( with._password ) ) )
					return true;
		return false;
	}


	public String toString()
	{
		if ( _description != null )
			return _description;
		else
		{
			String url;

			url = "jdbc:postgresql:";
			if ( _serverName != null )
			{
				if ( _portNumber == DEFAULT_PORT )
					url = url + "//" + _serverName + "/";
				else
					url = url + "//" + _serverName + ":" + _portNumber + "/";
			}
			else if ( _portNumber != DEFAULT_PORT )
				url = url + "//localhost:" + _portNumber + "/";
			if ( _databaseName != null )
				url = url + _databaseName;
			return "DataSource " + url;
		}
	}


	public synchronized Reference getReference()
	{
		Reference ref;

		// We use same object as factory.
		ref = new Reference( getClass().getName(), getClass().getName(), null );
		// Mandatory properties
		ref.add( new StringRefAddr( "description", _description ) );
		ref.add( new StringRefAddr( "databaseName", _databaseName ) );
		ref.add( new StringRefAddr( "loginTimeout", Integer.toString( _loginTimeout ) ) );
		// Optional properties
		if ( _user != null )
			ref.add( new StringRefAddr( "user", _user ) );
		if ( _password != null )
			ref.add( new StringRefAddr( "password", _password ) );
		if ( _serverName != null )
			ref.add( new StringRefAddr( "serverName", _serverName ) );
		if ( _portNumber != DEFAULT_PORT )
			ref.add( new StringRefAddr( "portNumber", Integer.toString( _portNumber ) ) );
		return ref;
	}
    

	public Object getObjectInstance( Object refObj, Name name, Context nameCtx, Hashtable env )
	throws NamingException
	{
		Reference ref;

		// Can only reconstruct from a reference.
		if ( refObj instanceof Reference )
		{
			ref = (Reference) refObj;
			// Make sure reference is of datasource class.
			if ( ref.getClassName().equals( getClass().getName() ) )
			{

				PostgresqlConnectionPoolDataSource ds;
				RefAddr addr;

				try
				{
					ds = (PostgresqlConnectionPoolDataSource) Class.forName( ref.getClassName() ).newInstance();
				}
				catch ( Exception except )
				{
					throw new NamingException( except.toString() );
				}
				// Mandatory properties
				ds._description = (String) ref.get( "description" ).getContent();
				ds._databaseName = (String) ref.get( "databaseName" ).getContent();
				ds._loginTimeout = Integer.parseInt( (String) ref.get( "loginTimeout" ).getContent() );
				// Optional properties
				addr = ref.get( "user" );
				if ( addr != null )
					ds._user = (String) addr.getContent();
				addr = ref.get( "password" );
				if ( addr != null )
					ds._password = (String) addr.getContent();
				addr = ref.get( "serverName" );
				if ( addr != null )
					ds._serverName = (String) addr.getContent();
				addr = ref.get( "portNumber" );
				if ( addr != null )
					ds._portNumber = Integer.parseInt( (String) addr.getContent() );
				addr = ref.get( "transactionTimeout" );
				return ds;

			}
			else
				throw new NamingException( "DataSource: Reference not constructed from class " + getClass().getName() );
		}
		else if ( refObj instanceof Remote )
			return refObj;
		else
			return null;
	}


    private transient ArrayList connectionsUnused = null;
    private transient ArrayList connectionsUsed = null;
    // protected internal methods
    protected void resetConnections()throws SQLException{
        print("resetConnections() Start Resetting Connections");
        if (connectionsUsed!=null){
            for (Iterator i=connectionsUsed.iterator();i.hasNext();){
                removeConnection((PostgresqlPooledConnection)i.next());
            }
        }
        if (connectionsUnused!=null){
            for (Iterator i=connectionsUnused.iterator();i.hasNext();){
                removeConnection((PostgresqlPooledConnection)i.next());
            }
        }
        connectionsUsed = new ArrayList();
        connectionsUnused = new ArrayList();
        for (int i=0;i<minNumberOfConnections;i++){
            PostgresqlPooledConnection ppc = _getPooledConnection();
            connectionsUnused.add(ppc);
        }
        print("resetConnections() Finished Resetting Connections");
    }

    protected void removeConnection(PostgresqlPooledConnection ppc){
        try {
            print("removeConnection() Removing connection: " + ppc);
            ppc.removeConnectionEventListener(this);
            ppc.callRealClose();
        } catch (SQLException e){
            ; // ignoore
        }
    }
    
    /**
     * Modified from the PostgresqlDataSource.getConnection();
     */
    protected PostgresqlPooledConnection _getPooledConnection() throws SQLException{
		java.sql.Connection conn;
		Properties info;
		String url;
        if ( _driver == null ){
			try{
				_driver = new org.postgresql.Driver();
                // set driver logger
			} catch ( SQLException except )	{
                // fix this to use error properties
                print( "_getPooledConnection() Failed to initialize JDBC driver: " + except );
				throw except;
			}
		}

		// Use info to supply properties that are not in the URL.
		info = new Properties();
		info.put( "loginTimeout", Integer.toString( _loginTimeout ) );

		// DriverManager will do that and not rely on the URL alone.
		if ( getUser() == null || getPassword() == null )
			throw new PSQLException( "postgresql.ds.userpswd" );
		info.put( "user", getUser() );
		info.put( "password", getPassword() );
        
		if ( _serverName != null )
			info.put( "PGHOST", _serverName );
		if ( _portNumber != DEFAULT_PORT )
			info.put( "PGPORT", Integer.toString( _portNumber ) );
		if ( _databaseName != null )
			info.put( "PGDBNAME", _databaseName );

		// Construct the URL suitable for this driver.
		url = "jdbc:postgresql:";

		// Attempt to establish a connection. Report a successful
		// attempt or a failure.
		try
		{
			conn = _driver.connect( url, info );
			// FIXME
			//		if ( ! ( conn instanceof postgresql.jdbc2.Connection ) ) {
			if ( ! ( conn instanceof org.postgresql.jdbc2.Connection ) )
			{
				//--------
                print( "_getPooledConnection() JDBC 1 connections not supported" );
				throw new PSQLException( "postgresql.ds.onlyjdbc2" );
			}
		}
		catch ( SQLException except )
		{
            print( "_getPooledConnection() getConnection failed " + except );
			throw except;
		}
		if ( conn != null){
			print( "_getPooledConnection() getConnection returning " + conn );
        }
        
		PostgresqlPooledConnection ppc =  new PostgresqlPooledConnection((org.postgresql.jdbc2.Connection)conn);
        ppc.addConnectionEventListener(this);
        print("_getPooledConnection() Creating new connection: " + ppc);
        return ppc;
    }

    private boolean hasChanged(String str1, String str2){
        if (str1 == str2){
            return false;
        }
        if (str1!=null){
            return str1.equals(str2);
        } else if (str2!=null) {
            return (str2.equals(str1));
        } else {
            // What? This makes no sense, right? Both are null, so the first
            // condition got them.  Weirdness.
            return true; // both are null. Log this
        }
    }

    protected void readProperties(){
        DefaultProperties dp = DefaultProperties.getDefaultProperties();
        minNumberOfConnections=dp.getIntProperty("psql.minNumberOfConnections",minNumberOfConnections);
        minNumberOfConnectionsUnused=dp.getIntProperty("psql.minNumberOfConnectionsUnused",minNumberOfConnectionsUnused);
        maxNumberOfConnections=dp.getIntProperty("psql.maxNumberOfConnections",maxNumberOfConnections);
        numberOfSecondsBeforeAutoClosed=dp.getIntProperty("psql.numberOfSecondsBeforeAutoClosed",numberOfSecondsBeforeAutoClosed);
        discardConnectionsWithExceptions=dp.getBooleanProperty("psql.discardConnectionsWithExceptions",discardConnectionsWithExceptions);
        _loginTimeout=dp.getIntProperty("psql.loginTimeout",_loginTimeout);
        _user = dp.getStringProperty("psql.user",_user);
        _password = dp.getStringProperty("psql.password",_password);
        _databaseName= dp.getStringProperty("psql.databaseName",_databaseName);
        _serverName = dp.getStringProperty("psql.serverName",_serverName);
        _portNumber= dp.getIntProperty("psql.portNumber",_portNumber);
    }

}
